//**************************************************************************
//   Copyright (C) 2004 by Petri Damstn
//   petri.damsten@iki.fi
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the
//   Free Software Foundation, Inc.,
//   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//**************************************************************************
#include "kmfmenupage.h"
#include "kmfbutton.h"
#include "kmfgeometry.h"
#include "kmfmenu.h"
#include "QMImage.h"
#include "templateobject.h"
#include <kmediafactory/projectinterface.h>
#include <kapplication.h>
#include <klocale.h>
#include <kdebug.h>
#include <kmessagebox.h>
#include <kprocess.h>
#include <qvariant.h>
#include <qobjectlist.h>
#include <qimage.h>
#include <qfile.h>
#include <qdir.h>
#include <qdatetime.h>
#include <sys/types.h>
#include <sys/stat.h>

KMFMenuPage::KMFMenuPage(QObject *parent, const char *name) :
  KMFWidget(parent, name), m_language("en"), m_modifiedLayers(0), m_titles(0),
  m_chapters(0), m_titleStart(0), m_chapterStart(0)
{
  m_buttons = new QPtrList<KMFButton>;
  m_buttons->setAutoDelete(true);
  setResolution(resolution(m_prjIf->type()));
}

KMFMenuPage::~KMFMenuPage()
{
  delete m_buttons;
}

void KMFMenuPage::fromXML(const QDomElement& element)
{
  KMFWidget::fromXML(element);
  m_titles = element.attribute("titles", "0").toInt();
  m_chapters = element.attribute("chapters", "0").toInt();
  QString s;
  if(m_titles > 0)
    s = QString("%1").arg((m_titleStart / m_titles) +1);
  else if(m_chapters > 0)
    s = QString("%1_%2").arg(m_titleStart+1)
                        .arg((m_chapterStart / m_chapters) +1);
  else if(m_chapters > 0)
    s = QString("%1_%2").arg(m_titleStart+1)
                        .arg(m_chapterStart+1);
  QString n = QString("%1_%2").arg(name()).arg(s);
  setName(n.ascii());
}

QSize KMFMenuPage::resolution(const QString &type)
{
  if(type == "DVD-PAL")
    return QSize(720, 576);
  else if(type == "DVD-NTSC")
    return QSize(720, 480);
  else
    return QSize(360, 240);
}

void KMFMenuPage::setResolution(QSize resolution)
{
  m_resolution = resolution;
  QString size =
      QString("%1x%2").arg(resolution.width()).arg(resolution.height());
  m_background.size(size.ascii());
  m_background.read("xc:#44444400");
  m_sub.size(size.ascii());
  m_sub.read("xc:#000000FF");
  m_subHighlight.size(size.ascii());
  m_subHighlight.read("xc:#000000FF");
  m_subSelect.size(size.ascii());
  m_subSelect.read("xc:#000000FF");
  m_temp.size(size.ascii());
  geometry().left().set(0, KMFUnit::Absolute);
  geometry().top().set(0, KMFUnit::Absolute);
  geometry().width().set(resolution.width(), KMFUnit::Absolute);
  geometry().height().set(resolution.height(), KMFUnit::Absolute);
}

bool KMFMenuPage::writeSpumuxXml(QDomDocument& doc)
{
  QDomElement root = doc.createElement("subpictures");
  QDomElement stream = doc.createElement("stream");
  QDomElement spu = doc.createElement("spu");

  spu.setAttribute("end", "00:00:00.00");
  if(m_modifiedLayers & Sub)
    spu.setAttribute("image", QString("%1_sub.png").arg(name()));
  if(m_modifiedLayers & Highlight)
    spu.setAttribute("highlight", QString("%1_highlight.png").arg(name()));
  if(m_modifiedLayers & Select)
    spu.setAttribute("select", QString("%1_select.png").arg(name()));
  spu.setAttribute("force", "yes");

  KMFButton *btn;
  // We can't search and loop in the same list concurrently
  QPtrList<KMFButton> temp(*m_buttons);
  for (btn = m_buttons->first(); btn; btn = m_buttons->next())
  {
    if(btn->isHidden())
      continue;
    // TODO: Button class should do this
    /*QDomComment comment =
        doc.createComment(QString("Button: %1").arg(btn->name()));
    spu.appendChild(comment);*/
    QDomElement button = doc.createElement("button");
    button.setAttribute("name", btn->name());
    QRect rc = btn->paintRect();
    button.setAttribute("x0", rc.left());
    button.setAttribute("y0", rc.top());
    button.setAttribute("x1", rc.right());
    button.setAttribute("y1", rc.bottom());
    if(btn->down())
      button.setAttribute("down", btn->down()->name());
    if(btn->up())
      button.setAttribute("up", btn->up()->name());
    if(btn->right())
      button.setAttribute("right", btn->right()->name());
    if(btn->left())
      button.setAttribute("left", btn->left()->name());
    spu.appendChild(button);
  }
  stream.appendChild(spu);
  root.appendChild(stream);
  doc.appendChild(root);
  return true;
}

bool KMFMenuPage::parseButtons()
{
  KMFButton* btn;

  for (btn = m_buttons->first(); btn; btn = m_buttons->next())
    if(btn->parseJump() == false)
      return false;
  for (btn = m_buttons->first(); btn; btn = m_buttons->next())
    btn->parseDirections();
  return true;
}

void KMFMenuPage::writeDvdAuthorXml(QDomElement& element)
{
  QDomDocument doc = element.ownerDocument();
  QDomElement pgc = doc.createElement("pgc");
  if(strcmp(name(), "vmgm") == 0)
    pgc.setAttribute("entry", "title");
  //else
    //pgc.setAttribute("entry", "ptt");
  pgc.setAttribute("pause", "0");
  QDomElement vob = doc.createElement("vob");
  QString file = QString("./menus/%1.mpg").arg(name());
  vob.setAttribute("file", file);
  vob.setAttribute("pause", "inf");
  pgc.appendChild(vob);

  KMFButton *btn;
  int i = 1;
  for (btn = m_buttons->first(); btn; btn = m_buttons->next(), ++i)
  {
    if(btn->isHidden())
      continue;
    QDomElement button = doc.createElement("button");
    button.setAttribute("name", btn->name());
    QDomText text = doc.createTextNode("");
    btn->writeDvdAuthorXml(text);
    button.appendChild(text);
    pgc.appendChild(button);
  }
  element.appendChild(pgc);
}

void KMFMenuPage::saveImages(QString projectPath)
{
  QString file;
  // Save subpicture files
  if(m_modifiedLayers & Sub)
  {
    //m_sub.quantizeColorSpace(Magick::TransparentColorspace);
    //m_sub.quantizeDither(true);
    //m_sub.quantizeColors(4);
    //m_sub.quantize();
    file = projectPath + QString("/menus/%1_sub.png").arg(name());
    m_sub.write(file.ascii());
  }

  if(m_modifiedLayers & Highlight)
  {
    //m_subHighlight.quantizeColorSpace(Magick::TransparentColorspace);
    //m_subHighlight.quantizeDither(true);
    //m_subHighlight.quantizeColors(4);
    //m_subHighlight.quantize();
    file = projectPath + QString("/menus/%1_highlight.png").arg(name());
    m_subHighlight.write(file.ascii());
  }

  if(m_modifiedLayers & Select)
  {
    //m_subSelect.quantizeColorSpace(Magick::TransparentColorspace);
    //m_subSelect.quantizeDither(true);
    //m_subSelect.quantizeColors(4);
    //m_subSelect.quantize();
    file = projectPath + QString("/menus/%1_select.png").arg(name());
    m_subSelect.write(file.ascii());
  }
  /*
  KMFEncoder e((projectPath +
  QString("menus/%1_bg.mpg").arg(name())).ascii());
  // TODO: NTSC
  double frameRate = 25.0;
  if(e.chooseFormat("vob") != omnividea::fobs::OkCode)
    return;
  if(e.setVideoParameters(m_resolution.width(), m_resolution.height(), 8000.0,
    frameRate, (char*)"mpeg2video") != omnividea::fobs::OkCode)
    return;
  if(e.setAudioParameters(48000, 2, 128.0, (char*)"mp2") !=
    omnividea::fobs::OkCode)
    return;
  QImage img = QMImage(m_background).image();
  e.open();
  e.addFramesAndSilence(img, (int)(frameRate + 0.5));
  e.close();
  */
  /*
  KMFEncoder e(projectPath + "menus/silence.mp2");
  if(e.chooseFormat("mp2") != omnividea::fobs::OkCode)
    return;
  if(e.setAudioParameters(48000, 2, 128.0, (char*)"mp2") !=
     omnividea::fobs::OkCode)
    return;
  e.open();
  e.addSilence((int)(25.0 + 0.5));
  e.close();
  */
  // TODO: Do this with ffmpeg when somebody knows how or/and ffmpeg can
  // make dvdauthor compatible mpeg2
  // Save as jpg now and run script to convert it to valid mpg file
  file = projectPath + QString("/menus/%1.jpg").arg(name());
  m_background.write(file.ascii());
  KProcess jpg2mpg;
  jpg2mpg.setWorkingDirectory(projectPath + "/menus/");
  //jpg2mpg.setUseShell(true, "/bin/sh");
  jpg2mpg << "sh" << "jpg2mpg.sh" << name();
  m_uiIf->logger()->connectProcess(&jpg2mpg);
  jpg2mpg.start(KProcess::Block, KProcess::AllOutput);
}

bool KMFMenuPage::writeSpumuxXml(QString projectPath)
{
  // Write spumux xml
  QString s;
  QDomDocument doc("");
  if(writeSpumuxXml(doc) == false)
    return false;

  QFile file(projectPath + "/menus/" + name() + ".xml");
  if (!file.open(IO_WriteOnly))
    return false;
  QTextStream stream(&file);
  stream << doc.toString();
  file.close();
  return true;
}

void KMFMenuPage::makeScript(QString projectPath)
{
  // TODO: This is for converting jpg to mpg. This should be done with ffmpeg
  // but I don't know how or if ffmpeg can do menu mpegs for dvdauthor
  QFile file(projectPath + "/menus/jpg2mpg.sh");
  if (!file.open(IO_WriteOnly))
    return;
  QTextStream stream(&file);
  stream << "#!/bin/sh\n\
  dd if=/dev/zero bs=4 count=48000 | toolame -b 128 -s 48 /dev/stdin $1.mp2\n\
  jpeg2yuv -n 1 -I p -f 1 -j $1.jpg | mpeg2enc -a 2 -f 8 -n p -F 3 -o $1.m2v\n\
  mplex -f 8 -o $1_bg.mpg $1.m2v $1.mp2\n\
  spumux $1.xml < $1_bg.mpg > $1.mpg\n\
  rm $1.mp2 $1.m2v $1_bg.mpg\n";
  file.close();
  chmod(file.name().ascii(), S_IRUSR|S_IXUSR);
}

bool KMFMenuPage::makeMpeg(QString projectPath)
{
  QString file;

  QDir dir(projectPath + "/menus/");
  QFileInfo script(projectPath + "/menus/jpg2mpg.sh");

  if(!dir.exists())
    dir.mkdir(dir.path());
  if(!script.exists())
    makeScript(projectPath);
  if(paint() == false)
    return false;
  if(writeSpumuxXml(projectPath) == false)
    return false;
  saveImages(projectPath);
  return true;
}

Magick::Image& KMFMenuPage::layer(Layer layer)
{
  switch(layer)
  {
    case Temp:
      return m_temp;
    case Sub:
      return m_sub;
    case Highlight:
      return m_subHighlight;
    case Select:
      return m_subSelect;
    case None:
    case Background:
    default:
      return m_background;
  }
  return m_background;
}

bool KMFMenuPage::paint()
{
  m_modifiedLayers = 0;
  return paintChildWidgets(this);
}

// Define this to draw rectangles around widgets
#define WIDGET_DEBUG 0

#define WIDGET_DEBUG_IMAGE  0
#define WIDGET_DEBUG_LABEL  1
#define WIDGET_DEBUG_FRAME  0
#define WIDGET_DEBUG_BUTTON 0
#define WIDGET_DEBUG_BOX    1
#define WIDGET_DEBUG_WIDGET 0

bool KMFMenuPage::paintChildWidgets(QObject* parent)
{
  const QObjectList *list = parent->children();
  QObject *obj;

  if(list)
  {
    for(QObjectListIt it(*list); (obj=it.current())!= 0; ++it)
    {
      if(obj->inherits("KMFWidget"))
      {
        KMFWidget* widget = static_cast<KMFWidget*>(obj);
        if(widget->isHidden())
          continue;
#if WIDGET_DEBUG
        QRect rc = widget->paintRect();
        QString type = widget->className();
        QString color;
        std::list<Magick::Drawable> drawList;

        kdDebug() << type << " (" << widget->name() << "): " << rc << endl;
        if(type == "KMFImage" && WIDGET_DEBUG_IMAGE)
          color = "blue";
        else if(type == "KMFLabel" && WIDGET_DEBUG_LABEL)
          color = "red";
        else if(type == "KMFFrame" && WIDGET_DEBUG_FRAME)
          color = "green";
        else if(type == "KMFButton" && WIDGET_DEBUG_BUTTON)
          color = "yellow";
        else if((type == "KMFVBox" || type == "KMFHBox")
                && WIDGET_DEBUG_BOX)
          color = "white";
        else if(type == "KMWidget" && WIDGET_DEBUG_WIDGET)
          color = "gray";
        if(!color.isEmpty())
        {
          drawList.push_back(Magick::DrawableFillColor(color.ascii()));
          drawList.push_back(Magick::DrawableFillOpacity(0.3));
          drawList.push_back(Magick::DrawableStrokeColor(color.ascii()));
          drawList.push_back(Magick::DrawableStrokeWidth(1));
          drawList.push_back(Magick::DrawableRectangle(rc.left(), rc.top(),
                                                      rc.right(), rc.bottom()));
          m_background.draw(drawList);
        }
#endif
        try
        {
          widget->paint(this);
        }
        catch(Magick::Exception &e)
        {
          kdDebug() << "Caught exception: " << endl << "  " << e.what() << endl
              <<  "  " << widget->name() << ": " << widget->paintRect() << endl;
          m_uiIf->message(KMF::Error,
              i18n("ImageMagick error: %1").arg(e.what()));
          return false;
        }
        m_modifiedLayers |= widget->layer();
        if(paintChildWidgets(obj) == false)
          return false;
        // Just to make ui more responsible
        if(m_uiIf->progress(0))
          return false;
      }
    }
  }
  return true;
}

KMFButton* KMFMenuPage::button(const QString& name)
{
  KMFButton *btn;

  // We can't search and loop in the same list concurrently
  QPtrList<KMFButton> temp(*m_buttons);
  for (btn = temp.first(); btn; btn = temp.next())
  {
    if(!btn->isHidden() && btn->name() == name)
      return btn;
  }
  return 0;
}

void KMFMenuPage::setProperty(const QString& name, QVariant value)
{
  KMFWidget::setProperty(name, value);
  if(name == "language")
  {
    m_language = value.toString();
    menu()->setLanguage(m_language);
  }
}

#include "kmfmenupage.moc"
