//**************************************************************************
//   Copyright (C) 2004 by Petri Damstn
//   petri.damsten@iki.fi
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the
//   Free Software Foundation, Inc.,
//   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//**************************************************************************
#include "videoobject.h"
#include "videoplugin.h"
#include "videooptions.h"
#include <ffmpeg_class.h>
#include <kapplication.h>
#include <kmimetype.h>
#include <kurl.h>
#include <kmdcodec.h>
#include <kstringhandler.h>
#include <klocale.h>
#include <kdebug.h>
#include <qdom.h>
#include <qimage.h>
#include <qcolor.h>
#include <qfileinfo.h>
#include <qdir.h>
#include <unistd.h>

static const char* aspectStrings[] = {"4:3", "16:9"};

static const char* hAlign[] = {"left", "center", "right", "default"};
static const char* vAlign[] = {"top", "center", "bottom", "default"};

KMFSubtitle::KMFSubtitle()
{
  font = KMF::Font(QFont());
  valign = KMF::Rect::Bottom;
  halign = KMF::Rect::Center;
}

VideoObject::VideoObject(QObject* parent)
  : MediaObject(parent, "video"), m_aspect(KMF::ASPECT_4_3)
{
  m_videoProperties = new KAction(
      i18n("&Properties"), "pencil", 0, this,
      SLOT(slotProperties()),plugin()->actionCollection(),"mob_properties");
  m_chapters.setAutoDelete(true);
  m_subtitles.setAutoDelete(true);
  connect(&m_decoder, SIGNAL(convertProgress(int)),
           this, SLOT(slotProgress(int)));
}

VideoObject::~VideoObject()
{
}

void VideoObject::actions(QPtrList<KAction>& actionList) const
{
  actionList.append(m_videoProperties);
}

void VideoObject::slotProgress(int progress)
{
  if(uiInterface()->setItemProgress(progress))
  {
    m_stopped = true;
    m_decoder.stop();
  }
}

void VideoObject::fromXML(const QDomElement& element)
{
  m_chapters.clear();
  m_audioStreams.clear();
  QDomNode n = element.firstChild();
  while(!n.isNull())
  {
    QDomElement e = n.toElement();
    if(!e.isNull())
    {
      if(e.tagName() == "video")
      {
        m_id = e.attribute("id");
        setTitle(e.attribute("title"));
        setPreviewURL(e.attribute("custom_preview"));
        QDomNode m = e.firstChild();
        while(!m.isNull())
        {
          QDomElement e2 = m.toElement();
          if(!e2.isNull())
          {
            if(e2.tagName() == "file")
            {
              m_decoder.addFile(e2.attribute("path"));
            }
            else if(e2.tagName() == "chapter")
            {
              KMFChapter* c = new KMFChapter;
              c->time = e2.attribute("time").toDouble();
              c->name = e2.attribute("name");
              m_chapters.append(c);
            }
            else if(e2.tagName() == "audio")
            {
              m_audioStreams.append(e2.attribute("language", "en"));
            }
            else if(e2.tagName() == "subtitle")
            {
              KMF::Font font;

              QDomNode m2 = e2.firstChild();
              while(!m2.isNull())
              {
                QDomElement e3 = m.toElement();
                if(!e3.isNull())
                {
                  if(e3.tagName() == "font")
                    font.fromXML(e3);
                }
                m2 = m2.nextSibling();
              }
              addSubtitle(e2.attribute("file"),
                  e2.attribute("language","en"),
                  (KMF::Rect::VAlign)e2.attribute("valign",
                      "2").toInt(),
                  (KMF::Rect::HAlign)e2.attribute("halign",
                      "1").toInt(),
                  font);
            }
          }
          m = m.nextSibling();
        }
        if(e.hasAttribute("auto_chapters"))
          setChapterSecs(e.attribute("auto_chapters").toDouble());
        setAspect((KMF::AspectRatio)e.attribute("aspect", "0").toInt());
      }
    }
    n = n.nextSibling();
  }
  // Older project files need these
  if(m_chapters.count() < 1)
    setChapterSecs(300.0);
  if(m_audioStreams.count() == 0)
  {
    while(m_audioStreams.count() < m_decoder.audioStreamCount())
      m_audioStreams.append("en");
  }
  if(m_id.isEmpty())
    generateId();
}

void VideoObject::toXML(QDomElement& element) const
{
  QDomDocument doc = element.ownerDocument();
  QDomElement video = doc.createElement("video");
  video.setAttribute("title", title());
  video.setAttribute("aspect", (int)m_aspect);
  video.setAttribute("id", m_id);

  if(m_previewURL.isValid())
    video.setAttribute("custom_preview", m_previewURL.prettyURL());

  QStringList lst = m_decoder.fileNames();
  for(QStringList::ConstIterator it = lst.begin(); it != lst.end(); ++it)
  {
    QDomElement e = doc.createElement("file");
    e.setAttribute("path", (*it));
    video.appendChild(e);
  }
  for(KMFChapterListIterator chapter(m_chapters); *chapter != 0; ++chapter)
  {
    QDomElement e = doc.createElement("chapter");
    e.setAttribute("name", (*chapter)->name);
    e.setAttribute("time", (*chapter)->time);
    video.appendChild(e);
  }
  for(QStringList::ConstIterator it = m_audioStreams.begin(); it !=
      m_audioStreams.end(); ++it)
  {
    QDomElement e = doc.createElement("audio");
    e.setAttribute("language", *it);
    video.appendChild(e);
  }
  for(KMFSubtitleListIterator subtitle(m_subtitles); *subtitle != 0; ++subtitle)
  {
    QDomElement e = doc.createElement("subtitle");
    e.setAttribute("language", (*subtitle)->language);
    e.setAttribute("file", (*subtitle)->file.prettyURL());
    e.setAttribute("halign", (int)(*subtitle)->halign);
    e.setAttribute("valign", (int)(*subtitle)->valign);
    QDomElement e2 = doc.createElement("font");
    (*subtitle)->font.toXML(e2);
    e.appendChild(e2);
    video.appendChild(e);
  }
  element.appendChild(video);
}

void VideoObject::writeDvdAuthorXml(QDomElement& element)
{
  QDir dir(projectInterface()->projectDir() + "/media/");

  QDomDocument doc = element.ownerDocument();
  QDomElement titles = doc.createElement("titles");

  QDomElement video = doc.createElement("video");
  video.setAttribute("aspect", aspectStrings[(int)m_aspect]);
  titles.appendChild(video);

  for(QStringList::ConstIterator it = m_audioStreams.begin(); it !=
      m_audioStreams.end(); ++it)
  {
    QDomElement audio = doc.createElement("audio");
    audio.setAttribute("lang", *it);
    titles.appendChild(audio);
  }

  QDomElement pgc = doc.createElement("pgc");
  for(QFFMPegFileList::const_iterator it = m_decoder.files().begin();
      it != m_decoder.files().end(); ++it)
  {
    QDomElement vob = doc.createElement("vob");

    if(m_subtitles.count() > 0)
      vob.setAttribute("file", dir.filePath(m_id + ".sub.mpg"));
    else if(strcmp((*it).fileFormat()->long_name, "MPEG PS format") != 0)
      vob.setAttribute("file", dir.filePath(m_id + ".mpg"));
    else
      vob.setAttribute("file", (*it).fileName());
    // TODO: How to do this for multiple files?
    if(it == m_decoder.files().begin())
    {
      QString chap("");
      for(KMFChapterListIterator chapter(m_chapters); *chapter != 0; ++chapter)
      {
        int n = (int)((*chapter)->time);
        if(!chap.isEmpty())
          chap += ",";
        QString s;
        chap += s.sprintf("%d:%.2d:%.2d", n/(60*60), (n%(60*60))/60, n%60);
      }
      vob.setAttribute("chapters", chap);
    }
    vob.setAttribute("pause", "0");
    pgc.appendChild(vob);
  }
  QDomElement post = doc.createElement("post");
  QDomText text = doc.createTextNode(" call vmgm menu entry title ; ");
  post.appendChild(text);
  pgc.appendChild(post);
  titles.appendChild(pgc);
  element.appendChild(titles);
}

int VideoObject::timeEstimate() const
{
  return 1;
}

bool VideoObject::convertToDVD(QString& fileName)
{
  bool result = true;
  QFileInfo fi(fileName);
  QDir dir(projectInterface()->projectDir() + "/media/");
  QFileInfo fic(dir.filePath(m_id + ".mpg"));

  m_stopped = false;

  if(!dir.exists())
    dir.mkdir(dir.path());

  if(fic.lastModified() < m_decoder.lastModified())
  {
    uiInterface()->message(KMF::Info,
        i18n("   Converting %1 to DVD format").arg(title()));
    uiInterface()->setItemTotalSteps(
        (int)((double)m_decoder.duration() * m_decoder.frameRate()));
    // ffmpeg -i koe.dv -hq -acodec ac3 -aspect 4:3 -pass 2 -target pal-dvd
    //        ./output.mpeg
    QFFMpegConvertTo dvd;
    if(projectInterface()->type() == "DVD-PAL")
      dvd["target"] = "pal-dvd";
    else
      dvd["target"] = "ntsc-dvd";
    dvd["hq"] = "";
    dvd["aspect"] = aspectStrings[(int)m_aspect];
    result = m_decoder.convertTo(dvd, fic.filePath());
    if(result == false)
    {
      if(m_stopped == false)
        uiInterface()->message(KMF::Error, i18n("   Conversion error."));
      QFile(fic.filePath()).remove();
    }
  }
  else
  {
    uiInterface()->message(KMF::Info,
        i18n("   Conversion seems to be up to date"));
  }
  fileName = fic.filePath();
  return result;

  /*
  omnividea::fobs::Transcoder
      transcoder(fileName, dir.filePath(fi.fileName() + ".mpg"));
  omnividea::fobs::ReturnCode error;

  error = transcoder.chooseFormat("vob");
  // NTSC
  error = transcoder.chooseVideoCodec(720, 576, 8000, 25.0, "mpeg2video");
  error = transcoder.chooseAudioCodec(48000, 2, 128, "mp2");
  error = transcoder.transcode();
  */
}

QString VideoObject::checkFontFile(const QString& file)
{
  QFileInfo fi(file);
  QDir dir("~/.spumux/");
  QFileInfo link(dir.filePath(fi.fileName()));

  if(!dir.exists())
    dir.mkdir(dir.path());
  if(!link.exists())
    symlink(file, link.filePath());
  return fi.fileName();
}

bool VideoObject::writeSpumuxXml(QDomDocument& doc)
{
  QDomElement root = doc.createElement("subpictures");
  QDomElement stream = doc.createElement("stream");
  QDomElement textsub = doc.createElement("textsub");

  // TODO: multiple subtitles
  textsub.setAttribute("filename", m_subtitles.at(0)->file.path());
  textsub.setAttribute("vertical-alignment",
                       vAlign[m_subtitles.at(0)->valign]);
  textsub.setAttribute("horizontal-alignment",
                       hAlign[m_subtitles.at(0)->halign]);
  if(!m_subtitles.at(0)->font.file().isEmpty())
    textsub.setAttribute("font", checkFontFile(m_subtitles.at(0)->font.file()));

  if(m_subtitles.at(0)->font.size() > 0)
    textsub.setAttribute("fontsize", m_subtitles.at(0)->font.size());

  stream.appendChild(textsub);
  root.appendChild(stream);
  doc.appendChild(root);
  return true;
}

bool VideoObject::writeSpumuxXml(const QString& fileName)
{
  // Write spumux xml
  QString s;
  QDomDocument doc("");
  if(writeSpumuxXml(doc) == false)
    return false;

  QFile file(fileName);
  if (!file.open(IO_WriteOnly))
    return false;
  QTextStream stream(&file);
  stream << doc.toString();
  file.close();
  return true;
}

void VideoObject::output(KProcess* process, char* buffer, int buflen)
{
  QRegExp re("[\n\r]");
  QRegExp bytes("INFO: (\\d+) bytes of data written");
  int n;

  m_buffer += QString::fromLatin1(buffer, buflen);
  while((n = m_buffer.find(re)) >= 0)
  {
    if(bytes.search(m_buffer.left(n)) > -1)
    {
      if(m_lastUpdate.elapsed() > 250)
      {
        if(uiInterface()->setItemProgress(bytes.cap(1).toInt() / 1024))
          process->kill();
        m_lastUpdate.start();
      }
    }
    ++n;
    m_buffer.remove(0, n);
  }
}

bool VideoObject::convertSubtitles(QString& fileName)
{
  QDir dir(projectInterface()->projectDir() + "/media/");
  QFileInfo fic(dir.filePath(m_id + ".sub.mpg"));
  QFileInfo fiXml(projectInterface()->projectDir() + "/media/" + id() + ".xml");
  QFileInfo fi(fileName);

  if(!dir.exists())
    dir.mkdir(dir.path());

  writeSpumuxXml(fiXml.filePath());

  if(!fic.exists() ||
     (fic.lastModified() < fi.lastModified() &&
      fiXml.lastModified() < fi.lastModified()))
  {
    uiInterface()->message(KMF::Info,
      i18n("   Adding subtitles for %1").arg(title()));

    //kdDebug() << KMF::Font::fileMap() << endl;

    KShellProcess m_spumux("bash");

    m_spumux << "spumux -P "
        + KShellProcess::quote(fiXml.filePath())
        + " < " + KShellProcess::quote(fileName)
        + " > " + KShellProcess::quote(fic.filePath());
    m_spumux.setWorkingDirectory(projectInterface()->projectDir() + "/media/");
    uiInterface()->logger()->connectProcess(&m_spumux,
        "INFO: \\d+ bytes of data written", KProcess::Stderr);
    connect(&m_spumux, SIGNAL(receivedStderr(KProcess*, char*, int)),
            this, SLOT(output(KProcess*, char*, int)));
    uiInterface()->setItemTotalSteps(fi.size()/1024);
    m_spumux.start(KProcess::Block, KProcess::Stderr);
    uiInterface()->setItemProgress(fi.size()/1024);
  }
  else
  {
    uiInterface()->message(KMF::Info,
    i18n("   Subtitle conversion seems to be up to date"));
  }
  fileName = fic.filePath();
  return true;
}

bool VideoObject::make()
{
  uiInterface()->message(KMF::Info,
      i18n("Preparing file(s) for %1").arg(title()));
  QString fileName = m_decoder.fileNames()[0];

  if(!m_decoder.isDVDCompatible() ||
      (m_subtitles.count() > 0 && m_decoder.files().count() > 0))
  {
    if(!convertToDVD(fileName))
      return false;
  }

  if(m_subtitles.count() > 0)
  {
    if(!convertSubtitles(fileName))
      return false;
  }

  uiInterface()->progress(1);
  return true;
}

void VideoObject::slotProperties()
{
  VideoOptions dlg(kapp->activeWindow());
  dlg.setData(*this);
  if (dlg.exec())
  {
    dlg.getData(*this);
    projectInterface()->setDirty();
  }
}

QPixmap VideoObject::pixmap() const
{
  return KMimeType::pixmapForURL(fileName());
}

#define PERCENT 1
#define LIGHT 50

bool VideoObject::isBlack(const QImage& img) const
{
  int pixelLimit = (img.height() * img.width() * PERCENT) / 100;
  int lightLimit = (255 * LIGHT) / 100;
  int pixel = 0;

  for (int y = 0; y < img.height(); y++)
  {
    for (int x = 0; x < img.width(); x++)
    {
      QRgb pix = img.pixel(x, y);
      if (qRed(pix) > lightLimit || qGreen(pix) > lightLimit ||
          qBlue(pix) > lightLimit)
        ++pixel;
      if(pixel > pixelLimit)
        return false;
    }
  }
  return true;
}

void VideoObject::addSubtitle(const KURL& file, const QString& language,
                              KMF::Rect::VAlign valign,
                              KMF::Rect::HAlign halign,
                              const KMF::Font& font)
{
  KMFSubtitle* sub = new KMFSubtitle;

  sub->file = file;
  sub->language = language;
  sub->valign = valign;
  sub->halign = halign;
  sub->font = font;

  m_subtitles.append(sub);
}

QImage VideoObject::getFrame(double time) const
{
  // TODO: Simplify this
  /*
  QDir dir(projectInterface()->projectDir() + "/media/");
  QString file;

  if(m_subtitles.count() > 0)
  {
    file = dir.filePath(m_id + ".sub.mpg");
  }
  else if(strcmp(m_decoder.files().first().fileFormat()->long_name,
                 "MPEG PS format") != 0)
  {
    file = dir.filePath(m_id + ".mpg");
  }

  if(!file.isEmpty() && QFileInfo(file).exists())
  {
    if(file != m_tmpDecoder.fileNames()[0])
    {
      m_tmpDecoder.clear();
      m_tmpDecoder.addFile(file);
    }
    return m_tmpDecoder.frame(QAVTime(time));
  }
  else*/
    return m_decoder.frame(QAVTime(time));
}

QImage VideoObject::preview(int chapter) const
{
  bool black = true;
  double time = 0.0;
  double startTime;
  QImage img;

  if (chapter == MainPreview)
  {
    if(m_previewURL.isValid())
    {
      img.load(m_previewURL.path());
      return img;
    }
  }
  else
  {
    const KMFChapter* c = m_chapters.at(chapter-1);
    time = c->time;
  }

  startTime = time;
  while(black && (time < startTime + 60.0))
  {
    img = getFrame(time);
    black = isBlack(img);
    if(black)
      kdDebug() << "Black frame: " << time << endl;
    time += 1.0;
  }
  return img;
}

QString VideoObject::text(int chapter) const
{
  if(chapter == MainTitle)
    return title();
  else
  {
    const KMFChapter* c = m_chapters.at(chapter-1);
    return c->name;
  }
}

void VideoObject::setChapterSecs(double secs)
{
  double d = duration();
  int chapters = (int)(d / secs) + 1;

  // Don not put chapter too close to end
  if(((double)chapters * secs) + 30.0 > d && chapters > 1)
    --chapters;
  m_chapters.clear();
  for(int i=0; i < chapters; ++i)
  {
    KMFChapter* c = new KMFChapter;
    c->time = i*secs;
    c->name = QString(i18n("Chapter %1")).arg(i + 1);
    m_chapters.append(c);
  }
}

int VideoObject::chapters() const
{
  return m_chapters.count();
}

void VideoObject::generateId()
{
  KMD5 md5;

  md5.update(m_decoder.fileNames().first());
  md5.update(QDateTime::currentDateTime().toString("ddMMyyyyhhmmsszzzap"));
  m_id = md5.hexDigest();
}

bool VideoObject::addFile(QString fileName)
{
  m_decoder.addFile(fileName);
  while(m_audioStreams.count() < m_decoder.audioStreamCount())
    m_audioStreams.append("en");
  setChapterSecs(300.0);
  if(m_id.isEmpty())
    generateId();
  return true;
}

void VideoObject::setTitleFromFileName()
{
  QString name = KURL(fileName()).fileName();
  int i;

  name.replace('-', ' ');
  name.replace('_', ' ');
  if (i = name.findRev('.'))
    name = name.left(i);
  name = KStringHandler::capwords(name);
  setTitle(name);
}

#include "videoobject.moc"
