//**************************************************************************
//   Copyright (C) 2004 by Petri Damstn
//   petri.damsten@iki.fi
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the
//   Free Software Foundation, Inc.,
//   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//**************************************************************************
#include "kmfproject.h"
#include "kmfapplication.h"
#include <ktempfile.h>
#include <kmessagebox.h>
#include <kio/netaccess.h>
#include <klocale.h>
#include <ksavefile.h>
#include <kstandarddirs.h>
#include <kdebug.h>
#include <qfile.h>
#include <qdir.h>

KMFProject::KMFProject(QObject *parent, const char *name) :
  QObject(parent, name), m_template(0), m_output(0), m_modified(false),
  m_loading(false)
{
  m_lastModified.setTime_t(0);
  m_directory = ::locateLocal("tmp", "kmf/");
  //kdDebug() << KStandardDirs::kde_default("tmp") << endl;
}

KMFProject::~KMFProject()
{
}

bool KMFProject::validProject() const
{
  setError();
  if(m_list.count() == 0)
  {
    setError(i18n("No media files."));
    return false;
  }
  if(!m_template)
  {
    setError(i18n("No template selected."));
    return false;
  }
  if(!m_output)
  {
    setError(i18n("No output type selected."));
    return false;
  }
  return true;
}

void KMFProject::addItem(KMF::MediaObject *mob)
{
  m_list.append(mob);
  emit newItem(mob);
  setDirty();
}

void KMFProject::removeItem(KMF::MediaObject *mob)
{
  m_list.remove(mob);
  emit itemRemoved(mob);
  setDirty();
}

void KMFProject::setType(const QString& type)
{
  if(type != m_type)
  {
    m_type = type;
    emit init(type);
  }
}

void KMFProject::setDirectory(const QString& directory)
{
  m_directory = directory;
  setDirty();
}

void KMFProject::setTitle(const QString& title)
{
  m_title = title;
  setDirty();
}

void KMFProject::setTemplateObj(KMF::TemplateObject* tmplate)
{
  m_template = tmplate;
  setDirty();
}

void KMFProject::setOutput(KMF::OutputObject* output)
{
  m_output = output;
  setDirty();
}

void KMFProject::fromXML(QString xml)
{
  QDomDocument doc("kmf_project");
  doc.setContent(xml);
  QDomElement element = doc.documentElement();
  m_type = element.attribute("type");
  m_directory = element.attribute("dir");
  m_title = element.attribute("title");
  m_lastModified = QDateTime::fromString(element.attribute("last_modified"),
                                         Qt::ISODate);
  emit init(m_type);

  QDomNode n = element.firstChild();
  while(!n.isNull())
  {
    QDomElement e = n.toElement();
    if(!e.isNull())
    {
      if(e.tagName() == "media")
      {
        KMF::Plugin *plugin =
          (KMF::Plugin *)kmfApp->pluginInterface()->
          child(e.attribute("plugin"));
        if(plugin)
        {
          KMF::MediaObject *mob = plugin->createMediaObject(e);
          if(mob)
            addItem(mob);
        }
      }
      if(e.tagName() == "template")
      {
        KMF::Plugin *plugin =
              static_cast<KMF::Plugin*>(kmfApp->pluginInterface()->
                                        child(e.attribute("plugin")));
        if(plugin)
        {
          m_template = static_cast<KMF::TemplateObject*>(
                       plugin->child(e.attribute("object")));
          if(m_template)
            m_template->fromXML(e);
        }
      }
      if(e.tagName() == "output")
      {
        KMF::Plugin *plugin =
              static_cast<KMF::Plugin*>(kmfApp->pluginInterface()->
                                        child(e.attribute("plugin")));
        if(plugin)
        {
          m_output = static_cast<KMF::OutputObject*>(
                       plugin->child(e.attribute("object")));
          if(m_output)
            m_output->fromXML(e);
        }
      }
    }
    n = n.nextSibling();
  }
}

void KMFProject::saveObj(QDomDocument& doc, QDomElement& root,
                         const char* name, KMF::Object* ob)
{
  if(ob)
  {
    QDomElement e = doc.createElement(name);
    e.setAttribute("plugin", ob->parent()->name());
    e.setAttribute("object", ob->name());
    ob->toXML(e);
    root.appendChild(e);
  }
}

QString KMFProject::toXML()
{
  KMF::MediaObject *obj;
  QDomDocument doc("kmf_project");
  doc.appendChild(doc.createProcessingInstruction("xml",
                  "version=\"1.0\" encoding=\"UTF-8\""));

  QDomElement root=doc.createElement("project");
  root.setAttribute("type", m_type);
  root.setAttribute("dir", m_directory);
  root.setAttribute("title", m_title);
  root.setAttribute("last_modified", m_lastModified.toString(Qt::ISODate));
  doc.appendChild(root);

  for(obj=m_list.first(); obj; obj=m_list.next())
    saveObj(doc, root, "media", obj);
  saveObj(doc, root, "template", m_template);
  saveObj(doc, root, "output", m_output);
  return doc.toString();
}

bool KMFProject::make()
{
  KMF::MediaObject *obj;

  if(!validProject())
    return false;

  QDir dir(m_directory);
  if(!dir.exists())
    dir.mkdir(dir.path());

 for(obj=m_list.first(); obj; obj=m_list.next())
    if(!obj->make())
      return false;
  if(!m_template->make())
    return false;
  if(!m_output->make())
    return false;
  return true;
}

int KMFProject::timeEstimate() const
{
  int result = 0;

  if(!validProject())
    return result;
  for(QPtrList<KMF::MediaObject>::ConstIterator obj = m_list.begin();
      obj != m_list.end(); ++obj)
    result += (*obj)->timeEstimate();
  result += m_template->timeEstimate();
  result += m_output->timeEstimate();
  return result;
}

bool KMFProject::open(const KURL &url)
{
  bool result = false;
  QString tmpFile;

  m_loading = true;
  if(KIO::NetAccess::download(url, tmpFile, kmfApp->mainWidget()))
  {
    QFile file(tmpFile);
    if(file.open(IO_ReadOnly))
    {
      QTextStream stream(&file);
      fromXML(stream.read());
      file.close();
      m_url = url;
      setDirty(false);
      result = true;
    }
    KIO::NetAccess::removeTempFile(tmpFile);
  }
  m_loading = false;
  return result;
}

bool KMFProject::save(KURL url)
{
  bool saveas = !url.isEmpty();

  if(url.isEmpty())
  {
    if(m_url.isEmpty())
      return false;
    else
      url = m_url;
  }

  if(KIO::NetAccess::exists(url, false, kmfApp->mainWidget()) && saveas)
  {
    if(KMessageBox::questionYesNo(kmfApp->activeWindow(),
        i18n("Do you want to overwrite %1").arg(url.prettyURL()))
      == KMessageBox::No)
    {
      return true;
    }
  }

  if(url.isLocalFile())
  {
    KSaveFile saveFile(url.path());

    if(saveFile.status() == 0)
      *(saveFile.textStream()) << toXML();
    else
      return false;
  }
  else
  {
    KTempFile tempFile;
    tempFile.setAutoDelete(true);
    if(tempFile.status() == 0)
    {
      *(tempFile.textStream()) << toXML();
      if(tempFile.close())
      {
        if (!KIO::NetAccess::upload(tempFile.name(), url, kmfApp->mainWidget()))
          return false;
      }
      else
        return false;
    }
    else
      return false;
  }
  m_url = url;
  setDirty(false);
  return true;
}

void KMFProject::setDirty(bool dirty)
{
  m_modified = dirty;
  if(dirty && !m_loading)
    m_lastModified = QDateTime::currentDateTime(Qt::LocalTime);
  emit modified(m_title, dirty);
}

#include "kmfproject.moc"
